#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2024 Dirk Sammel <dirk.sammel@gmail.com>
# SPDX-License-Identifier: BSD-2-Clause-Patent

from pydantic import BaseModel
from typing import List


class Message(BaseModel):
    message_header: str = ""
    create_sql: List[str] = []
    group_by: List[str] = []
    store_as: List[str] = []
    message_fields: List[str] = []


class SummaryMessage(Message):
    message_header: str = "APEL-summary-job-message: v0.3\n"

    create_sql: List[str] = [
        "Site TEXT NOT NULL",
        "Month INT NOT NULL",
        "Year INT NOT NULL",
        "StopTime INT NOT NULL",
        "WallDuration INT NOT NULL",
        "RecordID TEXT UNIQUE NOT NULL",
    ]

    group_by: List[str] = ["Site", "Month", "Year"]

    store_as: List[str] = [
        "COUNT(RecordID) as NumberOfJobs",
        "SUM(WallDuration) as WallDuration",
        "SUM(NormalisedWallDuration) as NormalisedWallDuration",
        "SUM(CpuDuration) as CpuDuration",
        "SUM(NormalisedCpuDuration) as NormalisedCpuDuration",
        "MIN(StopTime) as EarliestEndTime",
        "MAX(StopTime) as LatestEndTime",
    ]

    message_fields: List[str] = [
        "Site",
        "Month",
        "Year",
        "GlobalUserName",
        "VO",
        "VOGroup",
        "VORole",
        "SubmitHost",
        "Infrastructure",
        "NodeCount",
        "Processors",
        "EarliestEndTime",
        "LatestEndTime",
        "WallDuration",
        "CpuDuration",
        "NormalisedWallDuration",
        "NormalisedCpuDuration",
        "NumberOfJobs",
    ]


class SingleJobMessage(Message):
    message_header: str = "APEL-individual-job-message: v0.3\n"

    create_sql: List[str] = [
        "Site TEXT NOT NULL",
        "LocalJobId TEXT UNIQUE NOT NULL",
        "WallDuration INT NOT NULL",
        "StartTime INT NOT NULL",
        "EndTime INT NOT NULL",
    ]

    group_by: List[str] = [
        "Site",
        "LocalJobId",
        "WallDuration",
        "CpuDuration",
        "StartTime",
    ]

    store_as: List[str] = ["EndTime as EndTime"]

    message_fields: List[str] = [
        "Site",
        "SubmitHost",
        "MachineName",
        "Queue",
        "LocalJobId",
        "LocalUserId",
        "GlobalUserName",
        "FQAN",
        "VO",
        "VOGroup",
        "VORole",
        "WallDuration",
        "CpuDuration",
        "Processors",
        "NodeCount",
        "StartTime",
        "EndTime",
        "InfrastructureDescription",
        "InfrastructureType",
        "MemoryReal",
        "MemoryVirtual",
        "ServiceLevelType",
        "ServiceLevel",
    ]


class SyncMessage(Message):
    message_header: str = "APEL-sync-message: v0.1\n"

    create_sql: List[str] = [
        "Site TEXT NOT NULL",
        "Month INT NOT NULL",
        "Year INT NOT NULL",
        "SubmitHost TEXT NOT NULL",
        "RecordID TEXT UNIQUE NOT NULL",
    ]

    group_by: List[str] = ["Site", "Month", "Year", "SubmitHost"]

    store_as: List[str] = ["COUNT(RecordID) as NumberOfJobs"]

    message_fields: List[str] = ["Site", "SubmitHost", "NumberOfJobs", "Month", "Year"]



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Dirk Sammel <dirk.sammel@gmail.com>
# SPDX-License-Identifier: BSD-2-Clause-Patent

import logging
import sqlite3
from sqlite3 import Error
from datetime import datetime, timedelta, time, timezone
from time import sleep
import json
import sys
import requests
import urllib
from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.serialization import pkcs7
from pyauditor import Value, Operator, MetaOperator, MetaQuery, QueryBuilder, Record
from auditor_apel_plugin.config import Field, MessageType, Config
from auditor_apel_plugin.message import (
    SummaryMessage,
    SyncMessage,
    SingleJobMessage,
    Message,
)
from typing import Dict, List, Tuple

logger = logging.getLogger("apel_plugin")


def get_records(
    config: Config, client, start_time, delay_time, site=None, end_time=None
):
    sites_to_report = config.site.sites_to_report
    meta_key_site = config.auditor.site_meta_field

    site_ids = []

    if site is not None:
        site_ids = sites_to_report[site]
        logger.info(f"Getting records for site {site} with site_ids: {site_ids}")
    else:
        for k, v in sites_to_report.items():
            site_ids.extend(v)

        logger.info(
            f"Getting records for sites {list(sites_to_report.keys())} "
            f"with site_ids: {list(sites_to_report.values())}"
        )

    timeout_counter = 0
    records = []

    while timeout_counter < 2:
        try:
            start_time_value = Value.set_datetime(start_time)
            get_since_operator = Operator().gt(start_time_value)
            stop_time_query = QueryBuilder().with_stop_time(get_since_operator)
            if end_time is not None:
                end_time_value = Value.set_datetime(end_time)
                get_range_operator = get_since_operator.lt(end_time_value)
                stop_time_query = stop_time_query.with_stop_time(get_range_operator)
            for site in site_ids:
                site_operator = MetaOperator().contains(site)
                site_query = MetaQuery().meta_operator(meta_key_site, site_operator)
                query_string = stop_time_query.with_meta_query(site_query).build()
                records.extend(client.advanced_query(query_string))
            return records
        except Exception as e:
            if "timed" in str(e):
                timeout_counter += 1
                logger.warning(
                    f"Call to AUDITOR timed out {timeout_counter}/3! "
                    f"Trying again in {timeout_counter * delay_time}s"
                )
                sleep(timeout_counter * delay_time)
            else:
                logger.critical(e)
                raise

    logger.critical(
        "Call to AUDITOR timed out 3/3, quitting! "
        "Maybe increase auditor_timeout in the config"
    )

    sys.exit(1)


def get_begin_previous_month(current_time):
    first_current_month = current_time.replace(day=1)
    previous_month = first_current_month - timedelta(days=1)
    first_previous_month = previous_month.replace(day=1)
    begin_previous_month = datetime.combine(first_previous_month, time())
    begin_previous_month_utc = begin_previous_month.replace(tzinfo=timezone.utc)

    return begin_previous_month_utc


def get_begin_current_month(current_time):
    first_current_month = current_time.replace(day=1)
    begin_current_month = datetime.combine(first_current_month, time())
    begin_current_month_utc = begin_current_month.replace(tzinfo=timezone.utc)

    return begin_current_month_utc


def get_time_json(config):
    time_json_path = config.plugin.time_json_path

    try:
        with open(time_json_path, "r", encoding="utf-8") as f:
            time_dict = json.load(f)
    except FileNotFoundError:
        logger.warning(f"Path {time_json_path} not found, creating new time json")
        time_dict = create_time_json(time_json_path)

    return time_dict


def create_time_json(time_json_path):
    initial_report_time = datetime(1970, 1, 1, 0, 0, 0)
    time_dict = {
        "last_report_time": initial_report_time.isoformat(),
        "site_end_times": {},
    }

    try:
        with open(time_json_path, "w", encoding="utf-8") as f:
            json.dump(time_dict, f)
    except FileNotFoundError:
        logger.critical(f"Path {time_json_path} not found, could not create time json")
        raise

    return time_dict


def get_start_time(config, time_dict, site):
    try:
        start_time = datetime.fromisoformat(time_dict["site_end_times"][site])
    except KeyError:
        start_time = config.site.publish_since

    return start_time


def get_report_time(time_dict):
    report_time = datetime.fromisoformat(time_dict["last_report_time"])

    return report_time


def update_time_json(config, time_dict, site, stop_time, report_time):
    time_json_path = config.plugin.time_json_path

    time_dict["last_report_time"] = report_time.isoformat()
    time_dict["site_end_times"][site] = stop_time.isoformat()

    try:
        with open(time_json_path, "w", encoding="utf-8") as f:
            json.dump(time_dict, f)
    except FileNotFoundError:
        logger.critical(f"Path {time_json_path} not found, could not update time json")
        raise


def replace_record_string(string):
    updated_string = urllib.parse.unquote(string)

    return updated_string


def get_site_id(config, record):
    meta_key_site = config["auditor"]["meta_key_site"]

    try:
        site_id = record.meta.get(meta_key_site)[0]
        return site_id
    except AttributeError:
        logger.critical(f"No meta data found in {record.record_id}, aborting")
        raise
    except TypeError:
        logger.critical(f"No site name found in {record.record_id}, aborting")
        raise


def get_voms_info(config, record):
    meta_key_voms = config["auditor"]["meta_key_voms"]
    voms_dict = {}

    try:
        voms_string = replace_record_string(record.meta.get(meta_key_voms)[0])
    except TypeError:
        logger.warning(
            f"No VOMS information found in {record.record_id}, "
            "not sending VO, VOGroup, and VORole"
        )

        voms_dict["vo"] = None
        voms_dict["vogroup"] = None
        voms_dict["vorole"] = None

        return voms_dict

    if not voms_string.startswith("/"):
        logger.warning(
            f"VOMS information found in {record.record_id} has unknown "
            f"format: {voms_string}. Not sending VO, VOGroup, and VORole"
        )

        voms_dict["vo"] = None
        voms_dict["vogroup"] = None
        voms_dict["vorole"] = None

        return voms_dict

    voms_list = voms_string.split("/")
    voms_dict["vo"] = voms_list[1]

    if "Role" not in voms_string:
        logger.warning(
            f"No Role found in VOMS of {record.record_id}: {voms_string}, "
            "not sending VORole"
        )
        voms_dict["vorole"] = None

        if len(voms_list) == 2:
            voms_dict["vogroup"] = "/" + voms_list[1]
        else:
            voms_dict["vogroup"] = "/" + voms_list[1] + "/" + voms_list[2]
    elif len(voms_list) == 3:
        voms_dict["vogroup"] = "/" + voms_list[1]
        voms_dict["vorole"] = voms_list[2]
    else:
        voms_dict["vogroup"] = "/" + voms_list[1] + "/" + voms_list[2]
        voms_dict["vorole"] = voms_list[3]

    return voms_dict


def create_db(
    fields_dict: Dict[str, Field], message_type: MessageType
) -> sqlite3.Connection:
    message = Message()

    if message_type == MessageType.summaries:
        message = SummaryMessage()
    elif message_type == MessageType.individual_jobs:
        message = SingleJobMessage()
    elif message_type == MessageType.sync:
        message = SyncMessage()

    field_list = message.create_sql

    for k, v in fields_dict.items():
        if k not in message.message_fields:
            logger.critical(
                f"Field {k} not in list of possible fields: {message.message_fields}"
            )
            raise ValueError
        else:
            field_list.append(f"{k} NOT NULL")

    field_list_str = ",".join(field_list)

    create_db_str = "".join(
        ["CREATE TABLE IF NOT EXISTS records(", field_list_str, ")"]
    )

    conn = sqlite3.connect(":memory:")

    try:
        with conn:
            conn.execute(create_db_str)
    except Error as e:
        logger.critical(e)
        raise

    return conn


def fill_db(
    config: Config,
    conn: sqlite3.Connection,
    message_type: MessageType,
    fields_dict: Dict[str, Field],
    site: str,
    records: List[Record],
) -> sqlite3.Connection:
    message = Message()

    if message_type == MessageType.summaries:
        message = SummaryMessage()
    elif message_type == MessageType.individual_jobs:
        message = SingleJobMessage()
    elif message_type == MessageType.sync:
        message = SyncMessage()

    field_list = [field.split(" ")[0] for field in message.create_sql]

    for k in fields_dict.keys():
        field_list.append(k)

    field_list_str = ",".join(field_list)

    q_marks = ",".join(len(field_list) * ["?"])

    insert_db_str = "".join(
        ["INSERT INTO records(", field_list_str, ") VALUES(", q_marks, ")"]
    )

    for r in records:
        data_tuple = get_data_tuple(config, message_type, fields_dict, site, r)

        try:
            with conn:
                conn.execute(insert_db_str, data_tuple)
        except Error as e:
            logger.critical(e)
            raise

    return conn


def get_data_tuple(
    config: Config,
    message_type: MessageType,
    fields_dict: Dict[str, Field],
    site: str,
    record: Record,
) -> Tuple[int, float, str]:
    value_list = []

    if message_type == MessageType.summaries:
        month = record.stop_time.replace(tzinfo=timezone.utc).month
        year = record.stop_time.replace(tzinfo=timezone.utc).year
        stop_time = record.stop_time.replace(tzinfo=timezone.utc).timestamp()
        runtime = record.runtime
        record_id = record.record_id

        value_list = [site, month, year, stop_time, runtime, record_id]

    elif message_type == MessageType.individual_jobs:
        record_id = record.record_id
        runtime = record.runtime
        start_time = record.start_time.replace(tzinfo=timezone.utc).timestamp()
        stop_time = record.stop_time.replace(tzinfo=timezone.utc).timestamp()

        value_list = [site, record_id, runtime, start_time, stop_time]

    elif message_type == MessageType.sync:
        month = record.stop_time.replace(tzinfo=timezone.utc).month
        year = record.stop_time.replace(tzinfo=timezone.utc).year
        submithost_field = config.get_optional_fields().get("SubmitHost")
        if submithost_field is not None:
            submithost = replace_record_string(submithost_field.get_value(record))
        else:
            submithost = "None"

        record_id = record.record_id

        value_list = [site, month, year, submithost, record_id]

    for v in fields_dict.values():
        value = v.get_value(record)

        if isinstance(value, str):
            value = replace_record_string(value)

        value_list.append(value)

    data_tuple = tuple(value_list)

    return data_tuple


def group_db(
    conn: sqlite3.Connection, message_type: MessageType, fields_dict: Dict[str, Field]
) -> List[sqlite3.Row]:
    message = Message()

    if message_type == MessageType.summaries:
        message = SummaryMessage()
    elif message_type == MessageType.individual_jobs:
        message = SingleJobMessage()
    elif message_type == MessageType.sync:
        message = SyncMessage()

    group_by_list = message.group_by

    for k in fields_dict.keys():
        group_by_list.append(k)

    sql_group_by = ",".join(group_by_list)
    sql_store_as = ",".join(message.store_as)

    group_str = "".join(
        [
            "SELECT ",
            sql_group_by,
            ",",
            sql_store_as,
            " FROM records GROUP BY ",
            sql_group_by,
        ]
    )

    conn.row_factory = sqlite3.Row
    cur = conn.execute(group_str)
    grouped_sql = cur.fetchall()
    cur.close()
    conn.close()

    return grouped_sql


def create_message(message_type: MessageType, grouped_sql: List[sqlite3.Row]) -> str:
    message = Message()

    if message_type == MessageType.summaries:
        message = SummaryMessage()
    elif message_type == MessageType.individual_jobs:
        message = SingleJobMessage()
    elif message_type == MessageType.sync:
        message = SyncMessage()

    header = message.message_header
    message_fields = message.message_fields

    field_list = [header]

    for entry in grouped_sql:
        keys = entry.keys()

        for field in message_fields:
            if field in keys:
                field_list.append(f"{field}: {entry[field]}\n")
            else:
                field_list.append(f"{field}: None\n")

        field_list.append("%%\n")

    apel_message = "".join(field_list)

    return apel_message


def get_token(config):
    auth_url = config.authentication.auth_url
    client_cert = config.authentication.client_cert
    client_key = config.authentication.client_key
    verify_ca = config.authentication.verify_ca

    if verify_ca:
        ca_path = config.authentication.ca_path
    else:
        ca_path = False

    try:
        response = requests.get(
            auth_url, cert=(client_cert, client_key), verify=ca_path, timeout=10
        )
    except requests.Timeout:
        logger.critical("Timeout while getting token")
        raise

    token = response.json()["token"]

    return token


def sign_msg(config, msg):
    client_cert = config.authentication.client_cert
    client_key = config.authentication.client_key

    with open(client_cert, "rb") as cc:
        cert = x509.load_pem_x509_certificate(cc.read())

    with open(client_key, "rb") as ck:
        key = serialization.load_pem_private_key(ck.read(), None)

    options = [pkcs7.PKCS7Options.DetachedSignature, pkcs7.PKCS7Options.Text]

    signed_msg = (
        pkcs7.PKCS7SignatureBuilder()
        .set_data(bytes(msg, "utf-8"))
        .add_signer(cert, key, hashes.SHA256())
        .sign(serialization.Encoding.SMIME, options)
    )

    return signed_msg


def build_payload(msg):
    current_time = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S")
    empaid = f"{current_time[:8]}/{current_time}"

    payload = {"messages": [{"attributes": {"empaid": empaid}, "data": msg}]}

    return payload


def send_payload(config, token, payload):
    ams_url = config.authentication.ams_url
    verify_ca = config.authentication.verify_ca

    if verify_ca:
        ca_path = config.authentication.ca_path
    else:
        ca_path = False

    logger.debug(f"{ams_url}{token}")
    post = requests.post(
        f"{ams_url}{token}",
        json=payload,
        headers={"Content-Type": "application/json"},
        verify=ca_path,
    )

    return post


def convert_to_seconds(config, cpu_time):
    cpu_time_name = config["auditor"]["cpu_time_name"]
    cpu_time_unit = config["auditor"]["cpu_time_unit"]

    if cpu_time_unit == "seconds":
        return cpu_time
    elif cpu_time_unit == "milliseconds":
        return round(cpu_time / 1000)
    else:
        logger.critical(
            f"Unknown unit for {cpu_time_name}: {cpu_time_unit}. "
            "Possible values are seconds or milliseconds."
        )
        raise ValueError


def check_sites_in_records(config, records):
    sites_to_report = config["site"]["sites_to_report"]

    logger.debug(f"Sites to report from config: {list(sites_to_report.keys())}")

    sites_in_records = {get_site_id(config, r) for r in records}
    sites = []

    for site_id in sites_in_records:
        for k, v in sites_to_report.items():
            if site_id in v:
                sites.append(k)
                break

    logger.debug(f"Sites found in records: {sites}")

    return sites



import pytest
import yaml
from auditor_apel_plugin.config import (
    Config,
    Field,
    MetaField,
    ComponentField,
    ScoreField,
    NormalisedField,
    ConstantField,
    RecordField,
    PluginConfig,
    MessageType,
    get_loaders,
)
import pyauditor
from datetime import datetime, timezone
from pydantic import ValidationError
from pathlib import Path, PurePath

test_dir = PurePath(__file__).parent


class TestConfig:
    def test_plugin_config(self):
        with open(Path.joinpath(test_dir, "test_config.yml"), "r") as f:
            config: Config = yaml.load(f, Loader=get_loaders())

        log_level = "DEBUG"
        time_json_path = "time.json"
        report_interval = 10
        message_type = "summaries"

        plugin = PluginConfig(
            log_level=log_level,
            time_json_path=time_json_path,
            report_interval=report_interval,
            message_type=message_type,
        )

        assert plugin.log_level == "DEBUG"
        assert plugin.time_json_path == "time.json"
        assert plugin.report_interval == 10
        assert plugin.message_type is MessageType("summaries")

        field_config = config.get_field_config()

        value = field_config.mandatory["CpuDuration"].name

        assert value == "TotalCPU"

        mandatory_fields = config.get_mandatory_fields()

        value = mandatory_fields["CpuDuration"].name

        assert value == "TotalCPU"

        optional_fields = config.get_optional_fields()

        value = optional_fields["GlobalUserName"].name

        assert value == "subject"

        all_fields = config.get_all_fields()

        value = all_fields["VO"].name

        assert value == "voms"

        value = all_fields["CpuDuration"].name

        assert value == "TotalCPU"

        message_type = "something_else"

        with pytest.raises(Exception) as pytest_error:
            plugin = PluginConfig(
                log_level=log_level,
                time_json_path=time_json_path,
                report_interval=report_interval,
                message_type=message_type,
            )
        assert pytest_error.type is ValidationError

    def test_get_value_default(self):
        class TestField(Field):
            attribute: str

        test_field = TestField(attribute="test")

        record = pyauditor.Record(
            "record_id",
            datetime(1984, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
        )

        with pytest.raises(Exception) as pytest_error:
            test_field.get_value(record)

        assert pytest_error.type is NotImplementedError

    def test_get_value_meta_field(self):
        record = pyauditor.Record(
            "record_id",
            datetime(1984, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
        )

        meta = pyauditor.Meta()
        meta.insert("meta_test", ["value"])
        record.with_meta(meta)

        meta_field = MetaField(name="meta_test")
        value = meta_field.get_value(record)

        assert value == "value"

    def test_get_value_meta_field_fail(self):
        record = pyauditor.Record(
            "record_id",
            datetime(1984, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
        )

        meta = pyauditor.Meta()
        meta.insert("meta_test", ["value"])
        record.with_meta(meta)

        meta_field = MetaField(name="meta_test_missing")

        value = meta_field.get_value(record)

        assert value == "None"

        meta_field = MetaField(name="meta_test", function="missing_function")

        with pytest.raises(Exception) as pytest_error:
            meta_field.get_value(record)

        assert pytest_error.type is KeyError

    def test_meta_field_regex(self):
        record = pyauditor.Record(
            "record_id",
            datetime(1984, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
        )

        meta = pyauditor.Meta()
        meta.insert("meta_test", ["aaavaluebbb"])
        record.with_meta(meta)

        meta_field = MetaField(name="meta_test", regex=r"(?<=aaa).*?\S(?=bbb)")

        value = meta_field.get_value(record)

        assert value == "value"

        meta_field = MetaField(name="meta_test", regex=r"(?<=aaa).*?\S(?=ccc)")

        value = meta_field.get_value(record)

        assert value == "None"

    def test_get_value_component_field(self):
        record = pyauditor.Record(
            "record_id",
            datetime(1984, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
        )

        component = pyauditor.Component(name="test_component", amount=1000)
        record.with_component(component)

        component_field = ComponentField(name="test_component")
        value = component_field.get_value(record)

        assert value == 1000

        component_field = ComponentField(
            name="test_component",
            divide_by=1000,
        )
        value = component_field.get_value(record)

        assert value == 1

        component_field = ComponentField(name="test_component_2")

        with pytest.raises(Exception) as pytest_error:
            value = component_field.get_value(record)

        assert pytest_error.type is ValueError

    def test_get_value_score_field(self):
        record = pyauditor.Record(
            "record_id",
            datetime(1984, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
        )

        score = pyauditor.Score(name="test_score", value=2.5)
        component = pyauditor.Component(name="test_component", amount=1000)
        component.with_score(score)
        record.with_component(component)

        score_field = ScoreField(
            name="test_score",
            component_name="test_component",
        )
        value = score_field.get_value(record)

        assert value == 2.5

        score_field = ScoreField(
            name="test_score",
            component_name="test_component_2",
        )

        with pytest.raises(Exception) as pytest_error:
            value = score_field.get_value(record)

        assert pytest_error.type is ValueError

        score_field = ScoreField(
            name="test_score_2",
            component_name="test_component",
        )

        with pytest.raises(Exception) as pytest_error:
            value = score_field.get_value(record)

        assert pytest_error.type is ValueError

    def test_get_value_normalised_field(self):
        record = pyauditor.Record(
            "record_id",
            datetime(1984, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
        )

        score = pyauditor.Score(name="test_score", value=2.5)
        component = pyauditor.Component(name="test_component", amount=1000)
        component.with_score(score)
        record.with_component(component)

        score_field = ScoreField(
            name="test_score",
            component_name="test_component",
        )
        component_field = ComponentField(name="test_component")

        normalised_field = NormalisedField(
            base_value=component_field, score=score_field
        )
        value = normalised_field.get_value(record)

        assert value == 2500.0

    def test_get_value_normalised_field_fail(self):
        record = pyauditor.Record(
            "record_id",
            datetime(1984, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
        )

        score = pyauditor.Score(name="test_score", value=2.5)
        component = pyauditor.Component(name="test_component", amount=1000)
        component.with_score(score)
        record.with_component(component)

        record_field = RecordField(name="record_id")

        score_field = ScoreField(
            name="test_score",
            component_name="test_component",
        )

        normalised_field = NormalisedField(base_value=record_field, score=score_field)

        with pytest.raises(Exception) as pytest_error:
            normalised_field.get_value(record)

        assert pytest_error.type is TypeError

    def test_get_value_constant_field(self):
        constant_field = ConstantField(value="test")
        value = constant_field.get_value()

        assert value == "test"

        constant_field = ConstantField(value=15)
        value = constant_field.get_value()

        assert value == 15

    def test_get_value_record_field(self):
        record = pyauditor.Record(
            "record_id_123",
            datetime(1984, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
        )

        record_field = RecordField(name="record_id")

        value = record_field.get_value(record)

        assert value == "record_id_123"

        record_field = RecordField(name="start_time", modify="month")

        value = record_field.get_value(record)

        assert value == 3

    def test_get_value_record_field_fail(self):
        record = pyauditor.Record(
            "record_id_123",
            datetime(1984, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
        )

        record_field = RecordField(name="missing")

        with pytest.raises(Exception) as pytest_error:
            record_field.get_value(record)

        assert pytest_error.type is AttributeError

        record_field = RecordField(name="start_time", modify="missing")

        with pytest.raises(Exception) as pytest_error:
            record_field.get_value(record)

        assert pytest_error.type is AttributeError

    def test_loaders(self):
        test_yaml = """
                    !ComponentField
                      name: test_field
                    """

        config = yaml.load(test_yaml, Loader=get_loaders())
        value = config.name

        assert value == "test_field"



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Dirk Sammel <dirk.sammel@gmail.com>
# SPDX-License-Identifier: BSD-2-Clause-Patent

import logging
from logging import Logger
from pyauditor import AuditorClientBuilder
import argparse
from datetime import datetime
import yaml
import base64
import sys
from auditor_apel_plugin.core import (
    get_token,
    create_db,
    fill_db,
    group_db,
    create_message,
    sign_msg,
    build_payload,
    send_payload,
    get_records,
)
from auditor_apel_plugin.config import get_loaders, Config

TRACE = logging.DEBUG - 5


def run(logger: Logger, config, client, args):
    site = args.site

    message_type = config.plugin.message_type
    field_dict = config.get_all_fields()
    optional_fields = config.get_optional_fields()

    begin_date = datetime.fromisoformat(args.begin_date)
    end_date = datetime.fromisoformat(args.end_date)

    records = get_records(config, client, begin_date, 30, site, end_date)

    if len(records) == 0:
        logger.critical("No records found!")
        sys.exit(1)

    token = get_token(config)
    logger.debug(token)

    db = create_db(field_dict, message_type)
    filled_db = fill_db(config, db, message_type, field_dict, site, records)
    grouped_db = group_db(filled_db, message_type, optional_fields)
    message = create_message(message_type, grouped_db)
    logger.log(TRACE, message)
    signed_message = sign_msg(config, message)
    logger.log(TRACE, signed_message)
    encoded_message = base64.b64encode(signed_message).decode("utf-8")
    logger.log(TRACE, encoded_message)
    payload_message = build_payload(encoded_message)
    logger.log(TRACE, payload_message)
    post_message = send_payload(config, token, payload_message)
    logger.debug(post_message.status_code)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--begin-date",
        type=str,
        required=True,
        help="Begin of republishing (UTC): yyyy-mm-dd hh:mm:ss+00:00, "
        "e.g. 2023-11-27 13:31:10+00:00",
    )
    parser.add_argument(
        "--end-date",
        type=str,
        required=True,
        help="End of republishing (UTC): yyyy-mm-dd hh:mm:ss+00:00, "
        "e.g. 2023-11-29 21:10:54+00:00",
    )
    parser.add_argument(
        "-s", "--site", required=True, help="Site (GOCDB): UNI-FREIBURG, ..."
    )
    parser.add_argument("-c", "--config", required=True, help="Path to the config file")
    args = parser.parse_args()

    with open(args.config, "r") as f:
        config: Config = yaml.load(f, Loader=get_loaders())

    log_level = config.plugin.log_level
    log_format = (
        "[%(asctime)s] %(levelname)-8s %(message)s (%(pathname)s at line %(lineno)d)"
    )

    logging.addLevelName(TRACE, "TRACE")
    logging.basicConfig(
        level=log_level,
        format=log_format,
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    logging.getLogger("aiosqlite").setLevel("WARNING")
    logging.getLogger("urllib3").setLevel("WARNING")

    logger = logging.getLogger("apel_plugin")

    auditor_ip = config.auditor.ip
    auditor_port = config.auditor.port
    auditor_timeout = config.auditor.timeout

    builder = AuditorClientBuilder()
    builder = builder.address(auditor_ip, auditor_port).timeout(auditor_timeout)
    client = builder.build_blocking()

    try:
        run(logger, config, client, args)
    except KeyboardInterrupt:
        logger.critical("User abort")
    finally:
        logger.info("Republishing finished")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Dirk Sammel <dirk.sammel@gmail.com>
# SPDX-License-Identifier: BSD-2-Clause-Patent

import logging
from logging import Logger
from pyauditor import AuditorClientBuilder
from datetime import datetime, timedelta, timezone
import yaml
import argparse
import base64
from time import sleep
from auditor_apel_plugin.core import (
    get_token,
    get_time_json,
    get_report_time,
    get_start_time,
    sign_msg,
    build_payload,
    send_payload,
    update_time_json,
    get_begin_previous_month,
    get_begin_current_month,
    get_records,
    create_db,
    fill_db,
    group_db,
    create_message,
)
from auditor_apel_plugin.config import get_loaders, Config, MessageType

TRACE = logging.DEBUG - 5


def run(logger: Logger, config: Config, client):
    report_interval = config.plugin.report_interval
    sites_to_report = config.site.sites_to_report
    message_type = config.plugin.message_type
    field_dict = config.get_all_fields()
    optional_fields = config.get_optional_fields()

    token = get_token(config)
    logger.debug(token)

    while True:
        time_dict = get_time_json(config)
        last_report_time = get_report_time(time_dict)
        current_time = datetime.now()
        time_since_report = (current_time - last_report_time).total_seconds()

        if time_since_report < report_interval:
            logger.info("Not enough time since last report")
            sleep(report_interval - time_since_report)
            continue
        else:
            logger.info("Enough time since last report, create new report")

        if current_time.day < last_report_time.day:
            begin_month = get_begin_previous_month(current_time)
        else:
            begin_month = get_begin_current_month(current_time)

        for site in sites_to_report.keys():
            records = get_records(config, client, begin_month, 30, site=site)

            if len(records) == 0:
                logger.info(f"No new records for {site}")
                continue

            latest_stop_time = records[-1].stop_time.replace(tzinfo=timezone.utc)
            logger.debug(f"Latest stop time is {latest_stop_time}")

            sync_db = create_db({}, MessageType.sync)
            filled_sync_db = fill_db(
                config, sync_db, MessageType.sync, {}, site, records
            )
            grouped_sync_db = group_db(filled_sync_db, MessageType.sync, {})
            sync_message = create_message(MessageType.sync, grouped_sync_db)
            logger.debug(sync_message)
            signed_sync = sign_msg(config, sync_message)
            logger.debug(signed_sync)
            encoded_sync = base64.b64encode(signed_sync).decode("utf-8")
            logger.debug(encoded_sync)
            payload_sync = build_payload(encoded_sync)
            logger.debug(payload_sync)
            post_sync = send_payload(config, token, payload_sync)
            logger.debug(post_sync.status_code)

            if message_type == MessageType.individual_jobs:
                start_time = get_start_time(config, time_dict, site)
                logger.info(f"Getting records since {start_time}")
                records = [
                    r
                    for r in records
                    if r.stop_time.replace(tzinfo=timezone.utc) > start_time
                ]

                if len(records) == 0:
                    logger.info(f"No new records for {site}")
                    continue

            db = create_db(field_dict, message_type)
            filled_db = fill_db(config, db, message_type, field_dict, site, records)
            grouped_db = group_db(filled_db, message_type, optional_fields)
            message = create_message(message_type, grouped_db)
            logger.log(TRACE, message)
            signed_message = sign_msg(config, message)
            logger.log(TRACE, signed_message)
            encoded_message = base64.b64encode(signed_message).decode("utf-8")
            logger.log(TRACE, encoded_message)
            payload_message = build_payload(encoded_message)
            logger.log(TRACE, payload_message)
            post_message = send_payload(config, token, payload_message)
            logger.debug(post_message.status_code)

            latest_report_time = datetime.now()
            update_time_json(
                config, time_dict, site, latest_stop_time, latest_report_time
            )

        logger.info(
            "Next report scheduled for "
            f"{datetime.now() + timedelta(seconds=report_interval)}"
        )
        sleep(report_interval)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config", required=True, help="Path to the config file")
    args = parser.parse_args()

    with open(args.config, "r") as f:
        config: Config = yaml.load(f, Loader=get_loaders())

    log_level = config.plugin.log_level
    log_format = (
        "[%(asctime)s] %(levelname)-8s %(message)s (%(pathname)s at line %(lineno)d)"
    )

    logging.addLevelName(TRACE, "TRACE")
    logging.basicConfig(
        level=log_level,
        format=log_format,
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    logging.getLogger("aiosqlite").setLevel("WARNING")
    logging.getLogger("urllib3").setLevel("WARNING")

    logger = logging.getLogger("apel_plugin")

    auditor_ip = config.auditor.ip
    auditor_port = config.auditor.port
    auditor_timeout = config.auditor.timeout

    builder = AuditorClientBuilder()
    builder = builder.address(auditor_ip, auditor_port).timeout(auditor_timeout)
    client = builder.build_blocking()

    try:
        run(logger, config, client)
    except KeyboardInterrupt:
        logger.critical("User abort")
    finally:
        logger.critical("APEL plugin stopped")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2024 Dirk Sammel <dirk.sammel@gmail.com>
# SPDX-License-Identifier: BSD-2-Clause-Patent

import logging
import yaml
from datetime import datetime
from enum import Enum
from pydantic import BaseModel
from typing import Optional, Union, Dict, List, Callable
from pyauditor import Record
import re

logger = logging.getLogger("apel_plugin")


class MessageType(Enum):
    summaries = "summaries"
    individual_jobs = "individual_jobs"
    sync = "sync"


class Configurable(BaseModel):
    @classmethod
    def from_yaml(cls, loader: yaml.SafeLoader, node: yaml.nodes.MappingNode):
        return cls(**loader.construct_mapping(node, deep=True))


class PluginConfig(Configurable):
    log_level: str
    time_json_path: str
    report_interval: int
    message_type: MessageType


class SiteConfig(Configurable):
    publish_since: datetime
    sites_to_report: Dict[str, List[str]]


class AuditorConfig(Configurable):
    ip: str
    port: int
    timeout: int
    site_meta_field: str


class AuthConfig(Configurable):
    auth_url: str
    ams_url: str
    client_cert: str
    client_key: str
    ca_path: str
    verify_ca: bool


class Field(Configurable):
    def get_value(self, record: Optional[Record]) -> Union[str, int, float]:
        raise NotImplementedError()


class FieldConfig(Configurable):
    mandatory: Dict[str, Field]
    optional: Dict[str, Field] = {}


class ComponentField(Field):
    name: str
    divide_by: Optional[int] = None

    def get_value(self, record: Record) -> int:
        components = record.components
        value = None

        for c in components:
            if c.name == self.name:
                value = c.amount
                break

        if value is None:
            logger.critical(
                f"Component {self.name} not found in record {record.record_id}"
            )
            raise ValueError

        if self.divide_by is not None:
            value = round(value / self.divide_by)

        return value


class MetaField(Field):
    name: str
    regex: Optional[str] = None
    function: Optional[str] = None

    def get_value(self, record: Record) -> Union[str, int, float]:
        function_dict: Dict[str, Callable[[str], Union[str, int, float]]] = {}

        try:
            value = record.meta.get(self.name)[0]
        except TypeError:
            logger.warning(f"Meta {self.name} not found in record {record.record_id}")
            return "None"

        if self.regex is not None:
            re_match = re.search(self.regex, value)

            if re_match is not None:
                value = re_match.group(0)
                return value
            else:
                logger.warning(f"Pattern {self.regex} not found in {value}")
                return "None"
        elif self.function is not None:
            try:
                function = function_dict[self.function]
            except KeyError:
                logger.critical(
                    f"Function {self.function} not found in dictionary of allowed "
                    "functions"
                )
                raise
            value = function(value)
            return value

        return value


class ScoreField(Field):
    name: str
    component_name: str

    def get_value(self, record: Record) -> float:
        components = record.components
        scores = None
        value = None

        for c in components:
            if c.name == self.component_name:
                scores = c.scores
                break

        if scores is None:
            logger.critical(
                f"Component {self.name} not found in record {record.record_id}"
            )
            raise ValueError

        for s in scores:
            if s.name == self.name:
                value = s.value
                break

        if value is None:
            logger.critical(
                f"Score {self.name} not found in component {self.component_name} of "
                f"record {record.record_id}"
            )
            raise ValueError

        return value


class RecordField(Field):
    name: str
    modify: Optional[str] = None

    def get_value(self, record: Record) -> Union[str, int]:
        try:
            value = getattr(record, self.name)
        except AttributeError:
            logger.critical(
                f"Record {record.record_id} does not have attribute {self.name}"
            )
            raise

        if self.modify is not None:
            try:
                value = getattr(value, self.modify)
            except AttributeError:
                logger.critical(
                    f"Value {value} of type {type(value)} does not have attribute "
                    f"{self.name}"
                )
                raise

        return value


class NormalisedField(Field):
    base_value: Union[ComponentField, RecordField]
    score: ScoreField

    def get_value(self, record: Record) -> int:
        base_value = self.base_value.get_value(record)
        score_value = self.score.get_value(record)

        if isinstance(base_value, str):
            logger.critical(
                f"base_value of NormalisedField is a string: {base_value}. "
                "Multiplication not possible!"
            )
            raise TypeError
        value = round(base_value * score_value)

        return value


class NormalisedWallDurationField(NormalisedField):
    base_value: RecordField = RecordField(name="runtime")


class ConstantField(Field):
    value: Union[str, int, float]

    def get_value(self, record: Optional[Record] = None) -> Union[str, int, float]:
        return self.value


class Config(Configurable):
    plugin: PluginConfig
    site: SiteConfig
    auditor: AuditorConfig
    authentication: AuthConfig
    summary_fields: Optional[FieldConfig] = None
    individual_job_fields: Optional[FieldConfig] = None

    def get_field_config(self) -> FieldConfig:
        if self.plugin.message_type == MessageType.summaries:
            if self.summary_fields is not None:
                field_config = self.summary_fields
            else:
                logger.critical("summary_fields missing in config!")
                raise ValueError
        elif self.plugin.message_type == MessageType.individual_jobs:
            if self.individual_job_fields is not None:
                field_config = self.individual_job_fields
            else:
                logger.critical("individual_job_fields missing in config!")
                raise ValueError

        return field_config

    def get_mandatory_fields(self) -> Dict[str, Field]:
        return self.get_field_config().mandatory

    def get_optional_fields(self) -> Dict[str, Field]:
        return self.get_field_config().optional

    def get_all_fields(self) -> Dict[str, Field]:
        mandatory_dict = self.get_mandatory_fields()
        optional_dict = self.get_optional_fields()

        all_fields_dict = {**mandatory_dict, **optional_dict}

        return all_fields_dict


def get_loaders():
    loader = yaml.SafeLoader
    loader.add_constructor("!Config", Config.from_yaml)
    loader.add_constructor("!ComponentField", ComponentField.from_yaml)
    loader.add_constructor("!MetaField", MetaField.from_yaml)
    loader.add_constructor("!ScoreField", ScoreField.from_yaml)
    loader.add_constructor("!ConstantField", ConstantField.from_yaml)
    loader.add_constructor("!NormalisedField", NormalisedField.from_yaml)
    loader.add_constructor(
        "!NormalisedWallDurationField", NormalisedWallDurationField.from_yaml
    )
    loader.add_constructor("!RecordField", RecordField.from_yaml)
    return loader



import pytest
from auditor_apel_plugin.core import (
    get_begin_previous_month,
    get_begin_current_month,
    create_time_json,
    get_time_json,
    sign_msg,
    get_start_time,
    get_report_time,
    update_time_json,
    get_voms_info,
    replace_record_string,
    get_records,
    get_site_id,
    convert_to_seconds,
    check_sites_in_records,
)
from datetime import datetime, timezone
import os
import json
import subprocess
import pyauditor
from unittest.mock import patch, PropertyMock
from pathlib import Path, PurePath
from auditor_apel_plugin.config import Config, get_loaders
import yaml

test_dir = PurePath(__file__).parent


class FakeAuditorClient:
    def __init__(self, test_case=""):
        self.test_case = test_case

    def advanced_query(self, start_time):
        if self.test_case == "pass":
            return "good"
        if self.test_case == "fail_timeout":
            raise RuntimeError("Request timed out")
        if self.test_case == "fail_else":
            raise RuntimeError("Other RuntimeError")


def create_rec_metaless(rec_values, conf):
    rec = pyauditor.Record(rec_values["rec_id"], rec_values["start_time"])
    rec.with_stop_time(rec_values["stop_time"])
    rec.with_component(
        pyauditor.Component(conf["cores_name"], rec_values["n_cores"]).with_score(
            pyauditor.Score(conf["benchmark_name"], rec_values["hepscore"])
        )
    )
    rec.with_component(
        pyauditor.Component(conf["cpu_time_name"], rec_values["tot_cpu"])
    )
    rec.with_component(pyauditor.Component(conf["nnodes_name"], rec_values["n_nodes"]))

    return rec


def create_rec(rec_values, conf):
    rec = pyauditor.Record(rec_values["rec_id"], rec_values["start_time"])
    rec.with_stop_time(rec_values["stop_time"])
    rec.with_component(
        pyauditor.Component(conf["cores_name"], rec_values["n_cores"]).with_score(
            pyauditor.Score(conf["benchmark_name"], rec_values["hepscore"])
        )
    )
    rec.with_component(
        pyauditor.Component(conf["cpu_time_name"], rec_values["tot_cpu"])
    )
    rec.with_component(pyauditor.Component(conf["nnodes_name"], rec_values["n_nodes"]))
    meta = pyauditor.Meta()

    if rec_values["submit_host"] is not None:
        meta.insert(conf["meta_key_submithost"], [rec_values["submit_host"]])
    if rec_values["user_name"] is not None:
        meta.insert(conf["meta_key_username"], [rec_values["user_name"]])
    if rec_values["voms"] is not None:
        meta.insert(conf["meta_key_voms"], [rec_values["voms"]])
    if rec_values["site"] is not None:
        meta.insert(conf["meta_key_site"], [rec_values["site"]])
    rec.with_meta(meta)

    return rec


class TestAuditorApelPlugin:
    def test_get_begin_previous_month(self):
        time_a = datetime(2022, 10, 23, 12, 23, 55)
        time_b = datetime(1970, 1, 1, 00, 00, 00)

        result = get_begin_previous_month(time_a)
        assert result == datetime(2022, 9, 1, 00, 00, 00, tzinfo=timezone.utc)

        result = get_begin_previous_month(time_b)
        assert result == datetime(1969, 12, 1, 00, 00, 00, tzinfo=timezone.utc)

    def test_get_begin_current_month(self):

        time_a = datetime(2022, 10, 23, 12, 23, 55)
        time_b = datetime(1970, 1, 1, 00, 00, 00)

        result = get_begin_current_month(time_a)
        assert result == datetime(2022, 10, 1, 00, 00, 00, tzinfo=timezone.utc)

        result = get_begin_current_month(time_b)
        assert result == datetime(1970, 1, 1, 00, 00, 00, tzinfo=timezone.utc)

    def test_create_time_json(self):
        path = "/tmp/test.json"

        result = create_time_json(path)

        assert result["site_end_times"] == {}
        assert result["last_report_time"] == "1970-01-01T00:00:00"

        with open("/tmp/test.json", "r", encoding="utf-8") as f:
            result = json.load(f)

        os.remove(path)

        assert result["site_end_times"] == {}
        assert result["last_report_time"] == "1970-01-01T00:00:00"

    def test_create_time_json_fail(self):
        path = "/home/nonexistent/55/abc/time.db"

        with pytest.raises(Exception) as pytest_error:
            create_time_json(path)

        assert pytest_error.type is FileNotFoundError

    def test_get_time_json(self):
        with open(Path.joinpath(test_dir, "test_config.yml"), "r") as f:
            config: Config = yaml.load(f, Loader=get_loaders())

        path = "/tmp/nonexistent_55_abc_time.db"

        config.plugin.time_json_path = path

        result = get_time_json(config)
        os.remove(path)

        assert result["site_end_times"] == {}
        assert result["last_report_time"] == "1970-01-01T00:00:00"

        create_time_json(path)
        result = get_time_json(config)
        os.remove(path)

        assert result["site_end_times"] == {}
        assert result["last_report_time"] == "1970-01-01T00:00:00"

    def test_sign_msg(self):
        with open(Path.joinpath(test_dir, "test_config.yml"), "r") as f:
            config: Config = yaml.load(f, Loader=get_loaders())

        config.authentication.client_cert = Path.joinpath(test_dir, "test_cert.cert")
        config.authentication.client_key = Path.joinpath(test_dir, "test_key.key")

        result = sign_msg(config, "test")

        with open("/tmp/signed_msg.txt", "wb") as msg_file:
            msg_file.write(result)

        bashCommand = "openssl smime -verify -in /tmp/signed_msg.txt -noverify"
        process = subprocess.Popen(
            bashCommand.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        process.communicate()

        assert process.returncode == 0

    def test_sign_msg_fail(self):
        with open(Path.joinpath(test_dir, "test_config.yml"), "r") as f:
            config: Config = yaml.load(f, Loader=get_loaders())

        config.authentication.client_cert = "tests/nodir/test_cert.cert"
        config.authentication.client_key = "tests/no/dir/test_key.key"

        with pytest.raises(Exception) as pytest_error:
            sign_msg(config, "test")

        assert pytest_error.type is FileNotFoundError

        config.authentication.client_cert = str(
            Path.joinpath(test_dir, "test_cert.cert")
        )
        config.authentication.client_key = str(Path.joinpath(test_dir, "test_key.key"))

        result = sign_msg(config, "test")

        with open("/tmp/signed_msg.txt", "wb") as msg_file:
            msg_file.write(result.replace(b"test", b"TEST"))

        bashCommand = "openssl smime -verify -in /tmp/signed_msg.txt -noverify"
        process = subprocess.Popen(
            bashCommand.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        process.communicate()

        assert process.returncode == 4

    def test_get_start_time(self):
        with open(Path.joinpath(test_dir, "test_config.yml"), "r") as f:
            config: Config = yaml.load(f, Loader=get_loaders())

        path = "/tmp/test.json"

        config.plugin.time_json_path = path

        publish_since_list = [
            "1970-01-01 00:00:00+00:00",
            "2020-01-01 17:23:00+00:00",
        ]

        time_dict = create_time_json(path)

        for publish_since in publish_since_list:
            config.site.publish_since = datetime.fromisoformat(publish_since)
            result = get_start_time(config, time_dict, "TEST")
            time_dt = datetime.fromisoformat(publish_since)

            assert result == time_dt

        stop_time = datetime(1984, 3, 3, 0, 0, 0, tzinfo=timezone.utc)
        update_time_json(
            config,
            time_dict,
            "TEST",
            stop_time,
            datetime(1970, 1, 1, 0, 0, 0),
        )
        result = get_start_time(config, time_dict, "TEST")

        assert result == stop_time

        stop_time = datetime(2002, 12, 3, 0, 45, 0, tzinfo=timezone.utc)
        update_time_json(
            config,
            time_dict,
            "TEST-2",
            stop_time,
            datetime(1970, 1, 1, 0, 0, 0),
        )
        result = get_start_time(config, time_dict, "TEST-2")

        os.remove(path)

        assert result == stop_time

    def test_get_report_time(self):
        with open(Path.joinpath(test_dir, "test_config.yml"), "r") as f:
            config: Config = yaml.load(f, Loader=get_loaders())

        path = "/tmp/test.json"
        config.plugin.time_json_path = path

        time_dict = create_time_json(path)

        result = get_report_time(time_dict)

        initial_report_time = datetime(1970, 1, 1, 0, 0, 0)

        assert result == initial_report_time

        new_report_time = datetime(2023, 2, 10, 11, 13, 45)
        stop_time = datetime(2002, 12, 3, 0, 45, 0, tzinfo=timezone.utc)

        update_time_json(
            config,
            time_dict,
            "TEST-2",
            stop_time,
            new_report_time,
        )

        result = get_report_time(time_dict)
        os.remove(path)
        assert result == new_report_time

    def test_update_time_json(self):
        with open(Path.joinpath(test_dir, "test_config.yml"), "r") as f:
            config: Config = yaml.load(f, Loader=get_loaders())

        path = "/tmp/test.json"

        config.plugin.time_json_path = path

        time_dict = create_time_json(path)

        stop_time_list = [
            datetime(1984, 3, 3, 0, 0, 0),
            datetime(2022, 12, 23, 12, 44, 23),
            datetime(1999, 10, 1, 23, 17, 45),
        ]
        report_time_list = [
            datetime(1993, 4, 4, 0, 0, 0),
            datetime(2100, 8, 19, 14, 16, 11),
            datetime(1887, 2, 27, 0, 11, 31),
        ]
        site_list = ["SITE_A", "SITE_B"]

        for stop_time in stop_time_list:
            for report_time in report_time_list:
                for site in site_list:
                    update_time_json(config, time_dict, site, stop_time, report_time)

                    with open(path, "r", encoding="utf-8") as f:
                        time_dict_test = json.load(f)
                        last_end_time = time_dict_test["site_end_times"][site]
                        last_report_time = time_dict_test["last_report_time"]

                    assert last_end_time == stop_time.isoformat()
                    assert last_report_time == report_time.isoformat()

        os.remove(path)

    def test_update_time_json_fail(self):
        with open(Path.joinpath(test_dir, "test_config.yml"), "r") as f:
            config: Config = yaml.load(f, Loader=get_loaders())

        path = "/tmp/test.json"

        config.plugin.time_json_path = path

        time_dict = create_time_json(path)

        stop_time = datetime(1984, 3, 3, 0, 0, 0)
        report_time = datetime(1993, 4, 4, 0, 0, 0)
        site = "SITE_A"

        os.remove(path)

        path_new = "/dfghdfh/test.json"
        config.plugin.time_json_path = path_new

        with pytest.raises(Exception) as pytest_error:
            update_time_json(config, time_dict, site, stop_time, report_time)

        assert pytest_error.type is FileNotFoundError

    # def test_create_summary_db(self):
    #     sites_to_report = {
    #         "TEST_SITE_1": ["test-site-1"],
    #         "TEST_SITE_2": ["test-site-2"],
    #     }
    #     default_submit_host = "https://default.submit_host.de:1234/xxx"
    #     infrastructure_type = "grid"
    #     benchmark_name = "hepscore"
    #     cores_name = "Cores"
    #     cpu_time_name = "TotalCPU"
    #     cpu_time_unit = "seconds"
    #     nnodes_name = "NNodes"
    #     meta_key_site = "site_id"
    #     meta_key_submithost = "headnode"
    #     meta_key_voms = "voms"
    #     meta_key_username = "subject"
    #     benchmark_type = "hepscore23"

    #     conf = {}
    #     conf["site"] = {
    #         "sites_to_report": sites_to_report,
    #         "default_submit_host": default_submit_host,
    #         "infrastructure_type": infrastructure_type,
    #         "benchmark_type": benchmark_type,
    #     }
    #     conf["auditor"] = {
    #         "benchmark_name": benchmark_name,
    #         "cores_name": cores_name,
    #         "cpu_time_name": cpu_time_name,
    #         "cpu_time_unit": cpu_time_unit,
    #         "nnodes_name": nnodes_name,
    #         "meta_key_site": meta_key_site,
    #         "meta_key_submithost": meta_key_submithost,
    #         "meta_key_voms": meta_key_voms,
    #         "meta_key_username": meta_key_username,
    #     }

    #     runtime = 55

    #     rec_1_values = {
    #         "rec_id": "test_record_1",
    #         "start_time": datetime(1984, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
    #         "stop_time": datetime(1985, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
    #         "n_cores": 8,
    #         "hepscore": 10.0,
    #         "tot_cpu": 15520000,
    #         "n_nodes": 1,
    #         "site": "test-site-1",
    #         "submit_host": "https:%2F%2Ftest1.submit_host.de:1234%2Fxxx",
    #         "user_name": "%2FDC=ch%2FDC=cern%2FOU=Users%2FCN=test1: test1",
    #         "voms": "%2Fatlas%2Fde",
    #     }

    #     rec_2_values = {
    #         "rec_id": "test_record_2",
    #         "start_time": datetime(2023, 1, 1, 14, 24, 11).astimezone(tz=timezone.utc),
    #         "stop_time": datetime(2023, 1, 2, 7, 11, 45).astimezone(tz=timezone.utc),
    #         "n_cores": 1,
    #         "hepscore": 23.0,
    #         "tot_cpu": 12234325,
    #         "n_nodes": 2,
    #         "site": "test-site-2",
    #         "submit_host": "https:%2F%2Ftest2.submit_host.de:1234%2Fxxx",
    #         "user_name": "%2FDC=ch%2FDC=cern%2FOU=Users%2FCN=test2: test2",
    #         "voms": "%2Fatlas%2Fde",
    #     }

    #     rec_value_list = [rec_1_values, rec_2_values]
    #     records = []

    #     with patch(
    #         "pyauditor.Record.runtime", new_callable=PropertyMock
    #     ) as mocked_runtime:
    #         mocked_runtime.return_value = runtime

    #         for r_values in rec_value_list:
    #             rec = create_rec(r_values, conf["auditor"])
    #             records.append(rec)

    #         result = create_summary_db(conf, records)

    #     cur = result.cursor()

    #     cur.execute("SELECT * FROM records")
    #     content = cur.fetchall()

    #     cur.close()
    #     result.close()

    #     for idx, rec_values in enumerate(rec_value_list):
    #         assert content[idx][0] == list(sites_to_report.keys())[idx]
    #         assert content[idx][1] == replace_record_string(rec_values["submit_host"])
    #         assert (
    #             content[idx][2]
    #             == replace_record_string(rec_values["voms"]).split("/")[1]
    #         )
    #         assert content[idx][3] == replace_record_string(rec_values["voms"])
    #         assert content[idx][4] is None
    #         assert content[idx][5] == infrastructure_type
    #         assert content[idx][6] == rec_values["stop_time"].year
    #         assert content[idx][7] == rec_values["stop_time"].month
    #         assert content[idx][8] == rec_values["n_cores"]
    #         assert content[idx][9] == rec_values["n_nodes"]
    #         assert content[idx][10] == rec_values["rec_id"]
    #         assert content[idx][11] == runtime
    #         assert content[idx][12] == runtime * rec_values["hepscore"]
    #         assert content[idx][13] == rec_values["tot_cpu"]
    #         assert content[idx][14] == rec_values["tot_cpu"] * rec_values["hepscore"]
    #         assert (
    #             content[idx][15]
    #             == rec_values["start_time"].replace(tzinfo=timezone.utc).timestamp()
    #         )
    #         assert (
    #             content[idx][16]
    #             == rec_values["stop_time"].replace(tzinfo=timezone.utc).timestamp()
    #         )
    #         assert content[idx][17] == replace_record_string(rec_values["user_name"])

    #     rec_1_values["user_name"] = None
    #     records = []

    #     with patch(
    #         "pyauditor.Record.runtime", new_callable=PropertyMock
    #     ) as mocked_runtime:
    #         mocked_runtime.return_value = runtime

    #         for r_values in rec_value_list:
    #             rec = create_rec(r_values, conf["auditor"])
    #             records.append(rec)

    #         result = create_summary_db(conf, records)

    #     cur = result.cursor()

    #     cur.execute("SELECT * FROM records")
    #     content = cur.fetchall()

    #     assert content[0][17] is None

    #     cur.close()
    #     result.close()

    # def test_create_summary_db_fail(self):
    #     sites_to_report = {
    #         "TEST_SITE_1": ["test-site-1"],
    #         "TEST_SITE_2": ["test-site-2"],
    #     }
    #     default_submit_host = "https://default.submit_host.de:1234/xxx"
    #     infrastructure_type = "grid"
    #     benchmark_name = "hepscore"
    #     cores_name = "Cores"
    #     cpu_time_name = "TotalCPU"
    #     cpu_time_unit = "seconds"
    #     nnodes_name = "NNodes"
    #     meta_key_site = "site_id"
    #     meta_key_submithost = "headnode"
    #     meta_key_voms = "voms"
    #     meta_key_username = "subject"
    #     benchmark_type = "hepscore23"

    #     conf = {}
    #     conf["site"] = {
    #         "sites_to_report": sites_to_report,
    #         "default_submit_host": default_submit_host,
    #         "infrastructure_type": infrastructure_type,
    #         "benchmark_type": benchmark_type,
    #     }
    #     conf["auditor"] = {
    #         "benchmark_name": benchmark_name,
    #         "cores_name": cores_name,
    #         "cpu_time_name": cpu_time_name,
    #         "cpu_time_unit": cpu_time_unit,
    #         "nnodes_name": nnodes_name,
    #         "meta_key_site": meta_key_site,
    #         "meta_key_submithost": meta_key_submithost,
    #         "meta_key_voms": meta_key_voms,
    #         "meta_key_username": meta_key_username,
    #     }

    #     runtime = 55

    #     rec_1_values = {
    #         "rec_id": "test_record_1",
    #         "start_time": datetime(1984, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
    #         "stop_time": datetime(1985, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
    #         "n_cores": 8,
    #         "hepscore": 10.0,
    #         "tot_cpu": 15520000,
    #         "n_nodes": 1,
    #         "site": "test-site-1",
    #         "submit_host": "https:%2F%2Ftest1.submit_host.de:1234%2Fxxx",
    #         "user_name": "%2FDC=ch%2FDC=cern%2FOU=Users%2FCN=test1: test1",
    #         "voms": "%2Fatlas%2FRole=production",
    #     }

    #     rec_2_values = {
    #         "rec_id": "test_record_2",
    #         "start_time": datetime(2023, 1, 1, 14, 24, 11).astimezone(tz=timezone.utc),
    #         "stop_time": datetime(2023, 1, 2, 7, 11, 45).astimezone(tz=timezone.utc),
    #         "n_cores": 1,
    #         "hepscore": 23.0,
    #         "tot_cpu": 12234325,
    #         "n_nodes": 2,
    #         "site": "test-site-2",
    #         "submit_host": "https:%2F%2Ftest2.submit_host.de:1234%2Fxxx",
    #         "user_name": "%2FDC=ch%2FDC=cern%2FOU=Users%2FCN=test2: test2",
    #         "voms": "%2Fatlas",
    #     }

    #     rec_value_list = [rec_1_values, rec_2_values]
    #     records = []

    #     with patch(
    #         "pyauditor.Record.runtime", new_callable=PropertyMock
    #     ) as mocked_runtime:
    #         mocked_runtime.return_value = runtime

    #         for r_values in rec_value_list:
    #             rec = create_rec(r_values, conf["auditor"])
    #             records.append(rec)

    #         conf["auditor"]["cpu_time_name"] = "fail"
    #         with pytest.raises(Exception) as pytest_error:
    #             create_summary_db(conf, records)
    #         assert pytest_error.type is KeyError

    #         conf["auditor"]["cpu_time_name"] = "TotalCPU"
    #         conf["auditor"]["nnodes_name"] = "fail"
    #         with pytest.raises(Exception) as pytest_error:
    #             create_summary_db(conf, records)
    #         assert pytest_error.type is KeyError

    #         conf["auditor"]["nnodes_name"] = "NNodes"
    #         conf["auditor"]["cores_name"] = "fail"
    #         with pytest.raises(Exception) as pytest_error:
    #             create_summary_db(conf, records)
    #         assert pytest_error.type is KeyError

    #         conf["auditor"]["cores_name"] = "Cores"
    #         conf["auditor"]["benchmark_name"] = "fail"
    #         with pytest.raises(Exception) as pytest_error:
    #             create_summary_db(conf, records)
    #         assert pytest_error.type is KeyError

    #         rec_metaless = [create_rec_metaless(rec_1_values, conf["auditor"])]
    #         with pytest.raises(Exception) as pytest_error:
    #             create_summary_db(conf, rec_metaless)
    #         assert pytest_error.type is AttributeError

    def test_get_voms_info(self):
        default_submit_host = "https://default.submit_host.de:1234/xxx"
        benchmark_name = "hepscore"
        cores_name = "Cores"
        cpu_time_name = "TotalCPU"
        nnodes_name = "NNodes"
        meta_key_site = "site_id"
        meta_key_submithost = "headnode"
        meta_key_voms = "voms"
        meta_key_username = "subject"

        conf = {}
        conf["site"] = {
            "default_submit_host": default_submit_host,
        }
        conf["auditor"] = {
            "benchmark_name": benchmark_name,
            "cores_name": cores_name,
            "cpu_time_name": cpu_time_name,
            "nnodes_name": nnodes_name,
            "meta_key_site": meta_key_site,
            "meta_key_submithost": meta_key_submithost,
            "meta_key_voms": meta_key_voms,
            "meta_key_username": meta_key_username,
        }

        runtime = 55

        rec_1_values = {
            "rec_id": "test_record_1",
            "start_time": datetime(1984, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
            "stop_time": datetime(1985, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
            "n_cores": 8,
            "hepscore": 10.0,
            "tot_cpu": 15520000,
            "n_nodes": 1,
            "site": "test-site-1",
            "submit_host": "https:%2F%2Ftest1.submit_host.de:1234%2Fxxx",
            "user_name": "%2FDC=ch%2FDC=cern%2FOU=Users%2FCN=test1: test1",
            "voms": "%2Fatlas%2FRole=production",
        }

        rec_2_values = {
            "rec_id": "test_record_2",
            "start_time": datetime(2023, 1, 1, 14, 24, 11).astimezone(tz=timezone.utc),
            "stop_time": datetime(2023, 1, 2, 7, 11, 45).astimezone(tz=timezone.utc),
            "n_cores": 1,
            "hepscore": 23.0,
            "tot_cpu": 12234325,
            "n_nodes": 2,
            "site": "test-site-2",
            "submit_host": "https:%2F%2Ftest2.submit_host.de:1234%2Fxxx",
            "user_name": "%2FDC=ch%2FDC=cern%2FOU=Users%2FCN=test2: test2",
            "voms": "%2Fatlas",
        }

        rec_3_values = {
            "rec_id": "test_record_3",
            "start_time": datetime(2022, 1, 1, 14, 24, 11).astimezone(tz=timezone.utc),
            "stop_time": datetime(2023, 1, 2, 7, 11, 45).astimezone(tz=timezone.utc),
            "n_cores": 2,
            "hepscore": 3.0,
            "tot_cpu": 12265325,
            "n_nodes": 1,
            "site": "test-site-2",
            "user": "second_user",
            "submit_host": None,
            "user_name": "%2FDC=ch%2FDC=cern%2FOU=Users%2FCN=test2: test2",
            "voms": "%2Fatlas%2Fde%2FRole=production",
        }

        rec_4_values = {
            "rec_id": "test_record_4",
            "start_time": datetime(2022, 1, 1, 14, 24, 11).astimezone(tz=timezone.utc),
            "stop_time": datetime(2023, 1, 2, 7, 11, 45).astimezone(tz=timezone.utc),
            "n_cores": 2,
            "hepscore": 3.0,
            "tot_cpu": 12265325,
            "n_nodes": 1,
            "site": "test-site-2",
            "user": "second_user",
            "submit_host": None,
            "user_name": "%2FDC=ch%2FDC=cern%2FOU=Users%2FCN=test2: test2",
            "voms": "%2Fatlas%2Fde",
        }

        rec_5_values = {
            "rec_id": "test_record_4",
            "start_time": datetime(2022, 1, 1, 14, 24, 11).astimezone(tz=timezone.utc),
            "stop_time": datetime(2023, 1, 2, 7, 11, 45).astimezone(tz=timezone.utc),
            "n_cores": 2,
            "hepscore": 3.0,
            "tot_cpu": 12265325,
            "n_nodes": 1,
            "site": "test-site-2",
            "user": "second_user",
            "submit_host": None,
            "user_name": "%2FDC=ch%2FDC=cern%2FOU=Users%2FCN=test2: test2",
            "voms": None,
        }

        rec_6_values = {
            "rec_id": "test_record_4",
            "start_time": datetime(2022, 1, 1, 14, 24, 11).astimezone(tz=timezone.utc),
            "stop_time": datetime(2023, 1, 2, 7, 11, 45).astimezone(tz=timezone.utc),
            "n_cores": 2,
            "hepscore": 3.0,
            "tot_cpu": 12265325,
            "n_nodes": 1,
            "site": "test-site-2",
            "user": "second_user",
            "submit_host": None,
            "user_name": "%2FDC=ch%2FDC=cern%2FOU=Users%2FCN=test2: test2",
            "voms": "atlas",
        }

        rec_value_list = [
            rec_1_values,
            rec_2_values,
            rec_3_values,
            rec_4_values,
            rec_5_values,
            rec_6_values,
        ]
        records = []

        with patch(
            "pyauditor.Record.runtime", new_callable=PropertyMock
        ) as mocked_runtime:
            mocked_runtime.return_value = runtime

            for r_values in rec_value_list:
                rec = create_rec(r_values, conf["auditor"])
                records.append(rec)

        result = get_voms_info(conf, records[0])
        assert result["vo"] == "atlas"
        assert result["vogroup"] == "/atlas"
        assert result["vorole"] == "Role=production"

        result = get_voms_info(conf, records[1])
        assert result["vo"] == "atlas"
        assert result["vogroup"] == "/atlas"
        assert result["vorole"] is None

        result = get_voms_info(conf, records[2])
        assert result["vo"] == "atlas"
        assert result["vogroup"] == "/atlas/de"
        assert result["vorole"] == "Role=production"

        result = get_voms_info(conf, records[3])
        assert result["vo"] == "atlas"
        assert result["vogroup"] == "/atlas/de"
        assert result["vorole"] is None

        result = get_voms_info(conf, records[4])
        assert result["vo"] is None
        assert result["vogroup"] is None
        assert result["vorole"] is None

        result = get_voms_info(conf, records[5])
        assert result["vo"] is None
        assert result["vogroup"] is None
        assert result["vorole"] is None

    def test_replace_record_string(self):
        test_str_1 = "abcd"
        test_str_2 = "%2Fa%2Fb%2Fc%2Fd%2F"
        test_str_3 = "El%20Ni%C3%B1o"

        result = replace_record_string(test_str_1)
        assert result == "abcd"

        result = replace_record_string(test_str_2)
        assert result == "/a/b/c/d/"

        result = replace_record_string(test_str_3)
        assert result == "El Niño"

    def test_get_records(self):
        with open(Path.joinpath(test_dir, "test_config.yml"), "r") as f:
            config: Config = yaml.load(f, Loader=get_loaders())

        client = FakeAuditorClient("pass")

        sites_to_report = {"TEST_SITE_1": ["test-site-1"]}
        meta_key_site = "site_id"

        config.site.sites_to_report = sites_to_report
        config.auditor.site_meta_field = meta_key_site

        start_time_str = "2022-12-17 20:20:20+01:00"
        start_time = datetime.fromisoformat(start_time_str).replace(tzinfo=timezone.utc)

        result = get_records(config, client, start_time, 1)
        assert "".join(result) == "good"

    def test_get_records_fail(self):
        with open(Path.joinpath(test_dir, "test_config.yml"), "r") as f:
            config: Config = yaml.load(f, Loader=get_loaders())

        sites_to_report = {"TEST_SITE_1": ["test-site-1"]}
        meta_key_site = "site_id"

        config.site.sites_to_report = sites_to_report
        config.auditor.site_meta_field = meta_key_site

        start_time_str = "2022-12-17 20:20:20+01:00"
        start_time = datetime.fromisoformat(start_time_str).replace(tzinfo=timezone.utc)

        client = FakeAuditorClient("fail_timeout")

        with pytest.raises(SystemExit) as pytest_error:
            get_records(config, client, start_time, 1)
        assert pytest_error.type is SystemExit

        client = FakeAuditorClient("fail_else")

        with pytest.raises(Exception) as pytest_error:
            get_records(config, client, start_time, 1)
        assert pytest_error.type is RuntimeError

    def test_get_site_id(self):
        sites_to_report = (
            '{"TEST_SITE_1": ["test-site-1"], "TEST_SITE_2": ["test-site-2"]}'
        )
        default_submit_host = "https://default.submit_host.de:1234/xxx"
        infrastructure_type = "grid"
        benchmark_name = "hepscore"
        cores_name = "Cores"
        cpu_time_name = "TotalCPU"
        nnodes_name = "NNodes"
        meta_key_site = "site_id"
        meta_key_submithost = "headnode"
        meta_key_voms = "voms"
        meta_key_username = "subject"

        conf = {}
        conf["site"] = {
            "sites_to_report": sites_to_report,
            "default_submit_host": default_submit_host,
            "infrastructure_type": infrastructure_type,
        }
        conf["auditor"] = {
            "benchmark_name": benchmark_name,
            "cores_name": cores_name,
            "cpu_time_name": cpu_time_name,
            "nnodes_name": nnodes_name,
            "meta_key_site": meta_key_site,
            "meta_key_submithost": meta_key_submithost,
            "meta_key_voms": meta_key_voms,
            "meta_key_username": meta_key_username,
        }

        rec_1_values = {
            "rec_id": "test_record_1",
            "start_time": datetime(1984, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
            "stop_time": datetime(1985, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
            "n_cores": 8,
            "hepscore": 10.0,
            "tot_cpu": 15520000,
            "n_nodes": 1,
            "site": "test-site-1",
            "submit_host": "https:%2F%2Ftest1.submit_host.de:1234%2Fxxx",
            "user_name": "%2FDC=ch%2FDC=cern%2FOU=Users%2FCN=test1: test1",
            "voms": "%2Fatlas%2Fde",
        }

        rec_2_values = {
            "rec_id": "test_record_2",
            "start_time": datetime(2023, 1, 1, 14, 24, 11).astimezone(tz=timezone.utc),
            "stop_time": datetime(2023, 1, 2, 7, 11, 45).astimezone(tz=timezone.utc),
            "n_cores": 1,
            "hepscore": 23.0,
            "tot_cpu": 12234325,
            "n_nodes": 2,
            "site": "test-site-2",
            "submit_host": "https:%2F%2Ftest2.submit_host.de:1234%2Fxxx",
            "user_name": "%2FDC=ch%2FDC=cern%2FOU=Users%2FCN=test2: test2",
            "voms": "%2Fatlas%2Fde",
        }

        rec_1 = create_rec(rec_1_values, conf["auditor"])
        rec_2 = create_rec(rec_2_values, conf["auditor"])

        result = get_site_id(conf, rec_1)
        assert result == rec_1_values["site"]

        result = get_site_id(conf, rec_2)
        assert result == rec_2_values["site"]

    def test_get_site_id_fail(self):
        sites_to_report = (
            '{"TEST_SITE_1": ["test-site-1"], "TEST_SITE_2": ["test-site-2"]}'
        )
        default_submit_host = "https://default.submit_host.de:1234/xxx"
        infrastructure_type = "grid"
        benchmark_name = "hepscore"
        cores_name = "Cores"
        cpu_time_name = "TotalCPU"
        nnodes_name = "NNodes"
        meta_key_site = "site_id"
        meta_key_submithost = "headnode"
        meta_key_voms = "voms"
        meta_key_username = "subject"

        conf = {}
        conf["site"] = {
            "sites_to_report": sites_to_report,
            "default_submit_host": default_submit_host,
            "infrastructure_type": infrastructure_type,
        }
        conf["auditor"] = {
            "benchmark_name": benchmark_name,
            "cores_name": cores_name,
            "cpu_time_name": cpu_time_name,
            "nnodes_name": nnodes_name,
            "meta_key_site": meta_key_site,
            "meta_key_submithost": meta_key_submithost,
            "meta_key_voms": meta_key_voms,
            "meta_key_username": meta_key_username,
        }

        rec_1_values = {
            "rec_id": "test_record_1",
            "start_time": datetime(1984, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
            "stop_time": datetime(1985, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
            "n_cores": 8,
            "hepscore": 10.0,
            "tot_cpu": 15520000,
            "n_nodes": 1,
            "site": None,
            "submit_host": "https:%2F%2Ftest1.submit_host.de:1234%2Fxxx",
            "user_name": "%2FDC=ch%2FDC=cern%2FOU=Users%2FCN=test1: test1",
            "voms": "%2Fatlas%2Fde",
        }

        rec_2_values = {
            "rec_id": "test_record_2",
            "start_time": datetime(2023, 1, 1, 14, 24, 11).astimezone(tz=timezone.utc),
            "stop_time": datetime(2023, 1, 2, 7, 11, 45).astimezone(tz=timezone.utc),
            "n_cores": 1,
            "hepscore": 23.0,
            "tot_cpu": 12234325,
            "n_nodes": 2,
            "site": "test-site-2",
            "submit_host": "https:%2F%2Ftest2.submit_host.de:1234%2Fxxx",
            "user_name": "%2FDC=ch%2FDC=cern%2FOU=Users%2FCN=test2: test2",
            "voms": "%2Fatlas%2Fde",
        }

        rec_1 = create_rec(rec_1_values, conf["auditor"])
        rec_2 = create_rec_metaless(rec_2_values, conf["auditor"])

        with pytest.raises(Exception) as pytest_error:
            get_site_id(conf, rec_1)
        assert pytest_error.type is TypeError

        with pytest.raises(Exception) as pytest_error:
            get_site_id(conf, rec_2)
        assert pytest_error.type is AttributeError

    def test_convert_to_seconds(self):
        cpu_time_name = "TotalCPU"
        cpu_time_unit = "seconds"

        conf = {}
        conf["auditor"] = {
            "cpu_time_name": cpu_time_name,
            "cpu_time_unit": cpu_time_unit,
        }

        result = convert_to_seconds(conf, 1100)
        assert result == 1100

        result = convert_to_seconds(conf, 1500)
        assert result == 1500

        conf["auditor"]["cpu_time_unit"] = "milliseconds"

        result = convert_to_seconds(conf, 1100)
        assert result == 1

        result = convert_to_seconds(conf, 1500)
        assert result == 2

    def test_convert_to_seconds_fail(self):
        cpu_time_name = "TotalCPU"
        cpu_time_unit = "hours"

        conf = {}
        conf["auditor"] = {
            "cpu_time_name": cpu_time_name,
            "cpu_time_unit": cpu_time_unit,
        }

        with pytest.raises(Exception) as pytest_error:
            convert_to_seconds(conf, 1100)
        assert pytest_error.type is ValueError

    def test_check_sites_in_records(self):
        sites_to_report = {
            "TEST_SITE_1": ["test-site-1"],
            "TEST_SITE_2": ["test-site-2"],
        }
        benchmark_name = "hepscore"
        cores_name = "Cores"
        cpu_time_name = "TotalCPU"
        nnodes_name = "NNodes"
        meta_key_site = "site_id"
        meta_key_submithost = "headnode"
        meta_key_voms = "voms"
        meta_key_username = "subject"

        conf = {}
        conf["site"] = {
            "sites_to_report": sites_to_report,
        }
        conf["auditor"] = {
            "benchmark_name": benchmark_name,
            "cores_name": cores_name,
            "cpu_time_name": cpu_time_name,
            "nnodes_name": nnodes_name,
            "meta_key_site": meta_key_site,
            "meta_key_submithost": meta_key_submithost,
            "meta_key_voms": meta_key_voms,
            "meta_key_username": meta_key_username,
        }

        runtime = 55

        rec_1_values = {
            "rec_id": "test_record_1",
            "start_time": datetime(1984, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
            "stop_time": datetime(1985, 3, 3, 0, 0, 0).astimezone(tz=timezone.utc),
            "n_cores": 8,
            "hepscore": 10.0,
            "tot_cpu": 15520000,
            "n_nodes": 1,
            "site": "test-site-1",
            "submit_host": "https:%2F%2Ftest1.submit_host.de:1234%2Fxxx",
            "user_name": "%2FDC=ch%2FDC=cern%2FOU=Users%2FCN=test1: test1",
            "voms": "%2Fatlas%2Fde",
        }

        rec_2_values = {
            "rec_id": "test_record_2",
            "start_time": datetime(2023, 1, 1, 14, 24, 11).astimezone(tz=timezone.utc),
            "stop_time": datetime(2023, 1, 2, 7, 11, 45).astimezone(tz=timezone.utc),
            "n_cores": 1,
            "hepscore": 23.0,
            "tot_cpu": 12234325,
            "n_nodes": 2,
            "site": "test-site-2",
            "submit_host": "https:%2F%2Ftest2.submit_host.de:1234%2Fxxx",
            "user_name": "%2FDC=ch%2FDC=cern%2FOU=Users%2FCN=test2: test2",
            "voms": "%2Fatlas%2Fde",
        }

        rec_value_list = [rec_1_values, rec_2_values]
        records = []

        with patch(
            "pyauditor.Record.runtime", new_callable=PropertyMock
        ) as mocked_runtime:
            mocked_runtime.return_value = runtime

            for r_values in rec_value_list:
                rec = create_rec(r_values, conf["auditor"])
                records.append(rec)

            result = check_sites_in_records(conf, records)
            assert len(result) == 2

            conf["site"]["sites_to_report"] = {"TEST_SITE_1": ["test-site-1"]}

            result = check_sites_in_records(conf, records)
            assert len(result) == 1

            conf["site"]["sites_to_report"] = {"TEST_SITE_3": ["test-site-3"]}

            result = check_sites_in_records(conf, records)
            assert len(result) == 0
